/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: PKITest.java,v 1.17 2003/12/16 08:36:58 tk4 Exp $
 */
import java.io.*;
import java.math.BigInteger;
import java.security.*;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.sql.*;
import java.util.*;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.bouncycastle.jce.X509Principal;
import org.bouncycastle.jce.X509V3CertificateGenerator;
import org.ssonet.net.*;
import org.ssonet.net.impl.SSONETClientSocket;

import pki.*;

/**
 * Test class for for the PKI, which uses junit.
 *
 * @author Thomas Kriegelstein
 * @version $Id: PKITest.java,v 1.17 2003/12/16 08:36:58 tk4 Exp $
 */
public class PKITest extends TestCase {

	PKIDatabase database = null;
	PKICert certifier = null;
	Connection c = null;
	Properties settings = new Properties();

	/**
	 * Constructor
	 * @param name name of the test.
	 */
	public PKITest(String name) {
		super(name);
	}

	/**
	 * Preparation.
	 */
	public void setUp() {
		// set the test settings
		try {
			settings.load(new FileInputStream("data/pkiSettings.properties"));
		} catch (Exception e) {
			settings.setProperty(Main.keyStoreFileName, "data/pkikeys");
			settings.setProperty(Main.secretKeyPassword, "secret");
			settings.setProperty(Main.keyStorePassword, "secret");
			settings.setProperty(Main.dbType, "file");
			settings.setProperty(Main.fileDBFileName, "data/pkidata");
			settings.setProperty(Main.mysqlHost, "localhost");
			settings.setProperty(Main.mysqlPort, "3306");
			settings.setProperty(Main.mysqlDBName, "pki");
			settings.setProperty(Main.mysqlUsername, "pkiadmin");
			settings.setProperty(Main.mysqlPassword, "pkiadmin");
		}
		// initialize the modules
		database = new PKIDatabase(settings);
		certifier = new PKICert(settings);
		certifier.start();
		database.setCertifier(certifier);
		database.start();
		certifier.setDatabase(database);
	}

	/**
	 * Cleaning up.
	 */
	public void tearDown() {
		database.stop();
		database = null;
		certifier.stop();
		certifier = null;
	}

	/**
	 * Test 1: Keystore - PublicKey for "pki".
	 * @throws Exception in case of an error
	 */
	public void test1_Keystore() throws Exception {
		String filename = "data/pkikeys";
		KeyStore ks = KeyStore.getInstance("JKS");
		FileInputStream fis = new FileInputStream(filename);
		ks.load(fis, "secret".toCharArray());
		X509Certificate cert = (X509Certificate)ks.getCertificate("pki");
		assertTrue("cert!=null", cert != null);
	}

	/**
	 * Test 2: Database - read LastSerial.
	 * @throws Exception in case of error
	 */
	public void test2_Database() throws Exception {
		if (settings.getProperty(Main.dbType).equals("file")) {
			FileInputStream fis = new FileInputStream("data/pkidata");
			ObjectInputStream ois = new ObjectInputStream(fis);
			BigInteger l = (BigInteger)ois.readObject();
			Hashtable e = (Hashtable)ois.readObject();
			Hashtable r = (Hashtable)ois.readObject();
			Hashtable p = (Hashtable)ois.readObject();
			assertTrue(l != null);
			assertTrue(e != null);
			assertTrue(r != null);
			assertTrue(p != null);
			assertTrue("fis.available()==0", fis.available() == 0);
		} else {
			assertTrue("initilize MYSQL DB", initializeDB());
			// check, whether this database is initialized
			Statement s =
				c.createStatement(
					ResultSet.TYPE_SCROLL_INSENSITIVE,
					ResultSet.CONCUR_READ_ONLY);
			ResultSet rs = s.executeQuery("SELECT serial FROM persons");
			rs = s.executeQuery("SELECT person FROM roles");
			rs = s.executeQuery("SELECT person FROM pseudos");
			rs = s.executeQuery("SELECT name FROM config");
			rs =
				s.executeQuery(
					"SELECT value FROM config WHERE name='lastSerial'");
			assertTrue("lastSerial is set", rs.first());
		}
	}

	public boolean initializeDB() throws Exception {
		// initialize db driver
		java.lang.Class.forName("org.gjt.mm.mysql.Driver").newInstance();
		// connect to the database
		String host = settings.getProperty(Main.mysqlHost, "localhost");
		String port = settings.getProperty(Main.mysqlPort, "3306");
		c =
			DriverManager.getConnection(
				"jdbc:mysql://" + host + ":" + port + "/pki",
				"pkiadmin",
				"pkiadmin");
		return true;
	}

	/**
	 * Test 3: Certifier - issue a person pseudonym
	 * @throws Exception in case of error
	 */
	public void test3_Person() throws Exception {
		// lade PKI-Zertifikat aus KeyStore
		String filename = "data/pkikeys";
		KeyStore ks = KeyStore.getInstance("JKS");
		FileInputStream fis = new FileInputStream(filename);
		ks.load(fis, "secret".toCharArray());
		X509Certificate cert = (X509Certificate)ks.getCertificate("pki");
		assertTrue(certifier.testCertificate(cert) == true); //XXX
		// Erzeuge Personenpseudonym
		PublicKey pk = cert.getPublicKey();
		Properties p = new Properties();
		p.setProperty("CN", "Test");
		p.setProperty("C", "DE");
		X509Certificate pers =
			certifier.createPersonCertificate(pk, p, "secret".toCharArray());
		// teste, ob das Zertifikat korrekt ist
		byte[] enc1 = pers.getPublicKey().getEncoded();
		byte[] enc2 = pk.getEncoded();
		boolean same = enc1.length == enc2.length;
		assertTrue("same length", same);
		for (int i = 0; same && i < enc1.length; i++)
			same = enc1[i] == enc2[i];
		assertTrue("cert valid", certifier.testCertificate(pers));
		assertTrue(
			"correct serial",
			pers.getSerialNumber().equals(database.getLastSerial()));
		assertTrue("same key", same);
		// teste, ob das Zertifikat in die Datenbank eingetragen wurde
		X509Certificate lookup =
			database.getPersonCertificate(pers.getSerialNumber());
		assertTrue("found", lookup != null);
		assertTrue("equal", lookup.equals(pers));
	}

	/**
	 * Test 4: Certifier - issue a pseudonym for a person.
	 * @throws Exception in case of error
	 */
	public void test4_Pseudonym() throws Exception {
		X509Certificate oc = null;
		if (settings.getProperty(Main.dbType).equals("file")) {
			FileInputStream fis = new FileInputStream("data/pkidata");
			ObjectInputStream ois = new ObjectInputStream(fis);
			BigInteger l = (BigInteger)ois.readObject();
			Hashtable e = (Hashtable)ois.readObject();
			Hashtable r = (Hashtable)ois.readObject();
			Hashtable p = (Hashtable)ois.readObject();
			assertTrue(l != null);
			assertTrue(e != null);
			assertTrue(r != null);
			assertTrue(p != null);
			assertTrue("fis.available()==0", fis.available() == 0);
			oc = (X509Certificate)e.get(e.keys().nextElement());
		} else {
			// Lade ein zertifikat aus der persons-Tabelle
			if (c == null)
				assertTrue("initilize MYSQL DB", initializeDB());
			Statement s =
				c.createStatement(
					ResultSet.TYPE_SCROLL_INSENSITIVE,
					ResultSet.CONCUR_READ_ONLY);
			ResultSet rs = s.executeQuery("SELECT cert FROM persons");
			assertTrue("persons has entrys", rs.first());
			byte[] ser_cert = rs.getBytes(1);
			oc =
				(X509Certificate)CertificateFactory
					.getInstance("X.509", "BC")
					.generateCertificate(new ByteArrayInputStream(ser_cert));
			rs.close();
			s.close();
		}
		assertTrue("old cert", oc != null);
		// Erzeuge neues Pseudonym
		X509Certificate nc =
			certifier.createCertificate(
				oc,
				oc.getPublicKey(),
				"secret".toCharArray());
		// Teste korrektheit des neuen Pseudonyms
		assertTrue("new cert", nc != null);
		assertTrue("old != new", oc != nc);
		byte[] enc1 = oc.getPublicKey().getEncoded();
		byte[] enc2 = nc.getPublicKey().getEncoded();
		boolean same = enc1.length == enc2.length;
		assertTrue("same length", same);
		for (int i = 0; same && i < enc1.length; i++)
			same = enc1[i] == enc2[i];
		assertTrue("valid cert", certifier.testCertificate(nc));
		assertTrue(
			"correct serial",
			nc.getSerialNumber().equals(database.getLastSerial()));
		assertTrue("same key", same);
	}

	/**
	 * Test 5: Worker - Test zur Zusammenarbeit mit PKIWorker.
	 * @throws Exception im Fehlerfall
	 */
	/****** Faellt wegen doHandShake im Konstruktor des SSONETServerClientSocket aus
	public void test5_Worker() throws Exception {
	        FileInputStream   fis = new FileInputStream("data/pkidata");
	        ObjectInputStream ois = new ObjectInputStream(fis);
	        BigInteger l = (BigInteger)ois.readObject();
	        Hashtable  e = (Hashtable) ois.readObject();
	        Hashtable  r = (Hashtable) ois.readObject();
	        Hashtable  p = (Hashtable) ois.readObject();
	        assertTrue("fis.available()==0", fis.available()==0);
	        X509Certificate oc =
	                (X509Certificate)e.get(e.keys().nextElement());
	        assertTrue("old cert", oc!=null);
	        MySocket socket = new MySocket(oc.getEncoded(), oc);
	        PKIServer.PKIWorker worker =
	                new PKIServer.PKIWorker(certifier, database, "secret".toCharArray());
	        worker.set(socket);
	        worker.run();
	        InputStream is = socket.getResult();
	        int length = new DataInputStream(is).readInt();
	        byte[] carr = new byte[length];
	        assertTrue("read all", length == is.read(carr));
	        assertTrue("nothing left", is.available()==0);
	        // Zertifikat wiederherstellen und testen...
	        CertificateFactory cf =
	                CertificateFactory.getInstance("X.509");
	        X509Certificate c =
	                (X509Certificate)cf.generateCertificate(new ByteArrayInputStream(carr));
	        assertTrue("cert not null", c!=null);
	        assertTrue("valid cert", certifier.testCertificate(c));
	        // More assertions here? Done in Test 4 already?
	        //System.err.println("5 completed");
	}
	**/
	/**
	 * Spezielle Hilfsklasse fr Test 5.
	 */
	/**
	private static class MySocket extends SSONETServerClientSocket {
	
	        ByteArrayInputStream bais = null;
	        ByteArrayOutputStream baos = null;
	
	        public MySocket(byte[] arr, X509Certificate cert)
	                throws Exception
	        {
	                super(new SSONETContext(), new Socket());
	                ByteArrayOutputStream baos =
	                        new ByteArrayOutputStream();
	                new DataOutputStream(baos).writeInt(arr.length);
	                baos.write(arr);
	                this.bais = new ByteArrayInputStream(baos.toByteArray());
	                this.baos = new ByteArrayOutputStream();
	                getContext().setPartnerTestCertificate(cert);
	        }
	        public InputStream getInputStream() {
	                return bais;
	        }
	        public OutputStream getOutputStream() {
	                return baos;
	        }
	        public InputStream getResult() {
	                return new ByteArrayInputStream(baos.toByteArray());
	        }
	}
	**/
	/**
	 * Test 6: SSONET - Tests the cooperation with the SSONET library.
	 * @throws Exception in case of error
	 */
	public void test6_SSONET() throws Exception {
		// Erzeuge zwei DSA-Schlsselpaare
		KeyPairGenerator kpg1 = KeyPairGenerator.getInstance("DSA");
		assertTrue(kpg1 != null);
		kpg1.initialize(1024);
		KeyPair kp1 = kpg1.generateKeyPair();
		assertTrue(kp1 != null);
		KeyPair kp2 = kpg1.generateKeyPair();
		assertTrue(kp2 != null);

		// Erzeuge ein RSA-Schlsselpaar
		KeyPairGenerator kpg2 = KeyPairGenerator.getInstance("RSA");
		assertTrue(kpg2 != null);
		kpg2.initialize(1024);
		KeyPair kp3 = kpg2.generateKeyPair();
		assertTrue(kp3 != null);

		// Zertifikat generieren -> Personenpseudonym
		// Nicht wirklich benoetigt --
		//     aber der Vollstaendigkeit des Workflows wegen
		X509V3CertificateGenerator certGen = null;
		GregorianCalendar gregorianCalendar = null;

		certGen = new X509V3CertificateGenerator();
		certGen.setSubjectDN(new X509Principal("CN=Test1"));
		gregorianCalendar = (GregorianCalendar)Calendar.getInstance();
		certGen.setNotBefore(gregorianCalendar.getTime());
		gregorianCalendar.add(Calendar.YEAR, 1);
		certGen.setNotAfter(gregorianCalendar.getTime());
		certGen.setPublicKey(kp1.getPublic());
		certGen.setSignatureAlgorithm("SHA1withDSA");
		certGen.setIssuerDN(new X509Principal("CN=Test1"));
		certGen.setSerialNumber(BigInteger.ONE);
		X509Certificate cert1 =
			certGen.generateX509Certificate(kp1.getPrivate());
		assertTrue(cert1 != null);

		// zweites Zertifikat generieren -> Pseudonym
		certGen = new X509V3CertificateGenerator();
		certGen.setSubjectDN(new X509Principal("CN=Test2"));
		gregorianCalendar = (GregorianCalendar)Calendar.getInstance();
		certGen.setNotBefore(gregorianCalendar.getTime());
		gregorianCalendar.add(Calendar.YEAR, 1);
		certGen.setNotAfter(gregorianCalendar.getTime());
		certGen.setPublicKey(kp2.getPublic());
		certGen.setSignatureAlgorithm("SHA1withDSA");
		certGen.setIssuerDN(new X509Principal("CN=Test2"));
		certGen.setSerialNumber(BigInteger.valueOf(2));
		X509Certificate cert2 =
			certGen.generateX509Certificate(kp1.getPrivate());
		assertTrue(cert2 != null);

		// Personenpseudonym dazu erstellen (Test 3)
		Properties props = new Properties();
		props.setProperty("CN", "Test3");
		props.setProperty("C", "DE");
		char[] keypass = "secret".toCharArray();
		X509Certificate cert3 =
			certifier.createPersonCertificate(kp1.getPublic(), props, keypass);
		assertTrue(cert3 != null);
		assertTrue(certifier.testCertificate(cert3));

		// RSA-Cipherzertifikat generieren...
		certGen = new X509V3CertificateGenerator();
		certGen.setSubjectDN(new X509Principal("CN=Test2"));
		gregorianCalendar = (GregorianCalendar)Calendar.getInstance();
		certGen.setNotBefore(gregorianCalendar.getTime());
		gregorianCalendar.add(Calendar.YEAR, 1);
		certGen.setNotAfter(gregorianCalendar.getTime());
		certGen.setPublicKey(kp3.getPublic());
		certGen.setSignatureAlgorithm("SHA1withDSA");
		certGen.setIssuerDN(new X509Principal("CN=Test2"));
		certGen.setSerialNumber(BigInteger.valueOf(2));
		X509Certificate cert4 =
			certGen.generateX509Certificate(kp1.getPrivate());
		assertTrue(cert4 != null);

		// Context erzeugen
		SSONETContext result = new SSONETContext();
		// Context konfigurieren
		result.setBaseConfiguration(new BaseConfiguration());
		result.setOwnAnonServiceConfiguration(Boolean.FALSE);
		result.setOwnSigKey(kp1.getPrivate());
		result.setOwnTestCertificate(cert3);
		result.setOwnCiphKey(kp3.getPrivate());
		result.setOwnCipherCertificate(cert4);
		result.setOwnActionConfiguration(new ActionConfiguration(5, 5, 5, 5));
		result.addCertificateListener(new MyListener());

		// Server starten
		PKIServer serv = new PKIServer(settings);
		// Debug einschalten
		PKIServer.debug = true;
		PKIServer.PKIWorker.debug = true;
		//SSONETProtocolStream.debug = true;

		serv.setCertifier(certifier);
		serv.setDatabase(database);
		assertTrue(serv.start());

		// Verbindung aufbauen
		SSONETClientSocket socket =
			new SSONETClientSocket("127.0.0.1", 7777, result);
		InputStream is = socket.getInputStream();
		assertTrue(is != null);
		OutputStream os = socket.getOutputStream();
		assertTrue(os != null);

		// zweites Zertifikat bertragen
		System.out.println("bertrage zweites Zertifikat");
		byte[] carr1 = cert2.getEncoded();
		assertTrue(carr1 != null);
		(new DataOutputStream(os)).writeInt(carr1.length);
		os.write(carr1);
		os.flush();
		// Pseudonymzertifikat empfangen
		System.out.println("Warte auf Pseudonymzertifikat");
		assertTrue(is.available() >= 0);
		DataInputStream dis = new DataInputStream(is);
		int length = dis.readInt();
		byte[] carr2 = new byte[length];
		dis.readFully(carr2);
		System.out.println("Pseudonymzertifikat erhalten.");
		// Da kommt doch nix mehr? Protokollverletzung??
		assertTrue(is.available() == 0);
		// Schluss
		os.close();

		// alle Parameter des Resultats testen

		// Resultat auslesen
		CertificateFactory certFac = CertificateFactory.getInstance("X.509");
		ByteArrayInputStream bais = new ByteArrayInputStream(carr2);
		X509Certificate cert5 =
			(X509Certificate)certFac.generateCertificate(bais);
		// Wir wollen ein gueltiges Zertifikat zurueck...
		assertTrue("valid cert", certifier.testCertificate(cert5));
		// Das folgende funzt nur bei selber JVM...
		assertTrue(
			"correct serial",
			cert5.getSerialNumber().equals(database.getLastSerial()));
		// Die Schluessel sollen schon gleich sein...
		byte[] enc1 = cert2.getPublicKey().getEncoded();
		byte[] enc2 = cert5.getPublicKey().getEncoded();
		boolean same = enc1.length == enc2.length;
		assertTrue("same length", same);
		for (int i = 0; same && i < enc1.length; i++)
			same = enc1[i] == enc2[i];
		assertTrue("same key", same);
		// Wir wollen nicht das gesendete Zertifikat zurueckhaben..
		byte[] enc3 = cert2.getEncoded();
		byte[] enc4 = cert5.getEncoded();
		boolean diffrent = !(enc3.length == enc4.length);
		for (int i = 0; !diffrent && i < enc3.length; i++)
			diffrent = enc3[i] == enc4[i];
		assertTrue(diffrent);
	}

	/**
	 * Special help class for test 6: CertificateListener.
	 */
	private class MyListener implements CertificateListener {
		public int checkCertificate(SSONETContext ctx, X509Certificate cert) {
			try {
				System.out.println("Test-Checkcertificate called.");
				String filename = "data/pkikeys";
				KeyStore ks = KeyStore.getInstance("JKS");
				FileInputStream fis = new FileInputStream(filename);
				ks.load(fis, "secret".toCharArray());
				X509Certificate sigCert =
					(X509Certificate)ks.getCertificate("pki");
				// Dirty Hack: certs nochmal codieren und decodieren, damit die Formate gleich sind
				X509Certificate sigCert2 =
					(X509Certificate)CertificateFactory
						.getInstance("X.509", "BC")
						.generateCertificate(
							new ByteArrayInputStream(sigCert.getEncoded()));
				X509Certificate cert2 =
					(X509Certificate)CertificateFactory
						.getInstance("X.509", "BC")
						.generateCertificate(
							new ByteArrayInputStream(cert.getEncoded()));

				int result = sigCert2.equals(cert2) ? VALID : INVALID;

				System.out.println("Sigtest in PKITest: " + result);
				return result;
			} catch (Exception e) {
				return NOT_TESTED;
			}
		}
	}

	/**
	 * API-method for junit.
	 * @return the TestSuite containing the test in this class.
	 */
	public static Test suite() {
		return new TestSuite(PKITest.class);
	}
	public static void main(String[] args) {
		junit.textui.TestRunner.main(new String[] { "PKITest" });
	}
}
