/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: Main.java,v 1.14 2003/12/16 08:36:58 tk4 Exp $
 */
package pki;

import java.io.*;
import java.util.Properties;

/**
 * The main class does the evaluation of the command line parameters.
 *
 * @author Thomas Kriegelstein
 * @version $Id: Main.java,v 1.14 2003/12/16 08:36:58 tk4 Exp $
 */
public class Main {
	// standard values for system state
	private static boolean useGui = true;
	/** settings for configuration of the PKI */
	private static Properties settings = new Properties();
	public static final String keyStoreFileName = "keyStoreFileName";
	public static final String secretKeyPassword = "secretKeyPassword";
	public static final String keyStorePassword = "keyStorePassword";
	public static final String dbType = "dbType";
	public static final String fileDBFileName = "fileDBFileName";
	public static final String mysqlHost = "mysqlHost";
	public static final String mysqlPort = "mysqlPort";
	public static final String mysqlDBName = "mysqlDBName";
	public static final String mysqlUsername = "mysqlUsername";
	public static final String mysqlPassword = "mysqlPassword";
	private static String settingsFileName = "data/pkiSettings.properties";

	static PKIPanel panel = null;
	static PKIDatabase data = null;
	static PKICert cert = null;
	static PKIServer serv = null;

	/**
	 * Prints a usage message to standard out
	 */
	private static void usage() {
		System.out.println(
			"Usage: pki.Main [-keypass \"string\"] [-storephrase \"string\"] [-nogui | -test | -help]");
		String[] opts =
			{
				"Options are:",
				"  -keypass      %s  use secret for server operation",
				"  -storephrase  %s  use secret to validate the integrity of the keystore",
				"  -nogui            run without gui, default in headless mode",
				"  -mysql            use mysql database instead of file database",
				"  -initdb           initialize the database",
				"  -test             run the testsuite, test the installation",
				"  -help             print this help message and exit" };
		for (int i = 0; i < opts.length; i++)
			System.out.println(opts[i]);
	}

	/**
	 * Reads settings from file
	 */
	private static void loadSettings() {
		try {
			File sf = new File(settingsFileName);
			settings.load(new FileInputStream(sf));
		} catch (Exception e) {
			System.out.println(
				"Unable to read Settings from file: " + settingsFileName);
		}
	}

	/**
	 * Stores the settings into file
	 */
	public static void saveSettings() {
		System.out.println("Main.saveSettings called.");
		try {
			File sf = new File(settingsFileName);
			settings.store(new FileOutputStream(sf), "PKI Server settings");
		} catch (Exception e) {
			System.out.println("Cannot write settings to File");
			e.printStackTrace();
		}
	}

	/**
	 * Sets default settings for the PKI Server
	 */
	private static void setDefaultSettings() {
		settings = new Properties();
		settings.setProperty(keyStoreFileName, "data/pkikeys");
		settings.setProperty(secretKeyPassword, "secret");
		settings.setProperty(keyStorePassword, "secret");
		settings.setProperty(dbType, "mysql");
		settings.setProperty(fileDBFileName, "data/pkidata");
		settings.setProperty(mysqlHost, "localhost");
		settings.setProperty(mysqlPort, "3306");
		settings.setProperty(mysqlDBName, "pki");
		settings.setProperty(mysqlUsername, "pkiadmin");
		settings.setProperty(mysqlPassword, "pkiadmin");
	}

	/**
	 * Main method. Handles commandline options.
	 */
	public static void main(String[] args) throws Exception {
		//System.out.println("Main.main called.");
		setDefaultSettings();
		// load settings from file
		loadSettings();
		// get command line options
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("?") || args[i].equals("-help")) {
				usage();
				System.exit(0);
				break;
			} else if (args[i].equals("-storephrase")) {
				i += 1;
				if (i < args.length) {
					settings.setProperty(keyStorePassword, args[i]);
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else if (args[i].equals("-keypass")) {
				i += 1;
				if (i < args.length) {
					settings.setProperty(secretKeyPassword, args[i]);
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else if (args[i].equals("-test")) {
				System.out.println("Running tests on PKI.");
				String[] opts = { "PKITest" };
				junit.textui.TestRunner.main(opts);
				System.exit(0);
				break;
			} else if (args[i].equals("-nogui") || args[i].equals("-server")) {
				useGui = false;
			} else if (args[i].equals("-mysql")) {
				settings.setProperty(dbType, "mysql");
			} else if (args[i].equals("-initdb")) {
				initDBCommandline();
				System.exit(0);
			} else {
				usage();
				System.exit(0);
				break;
			}
		}
		// check if gui can really be used
		if (useGui) { // not disabled
			try {
				java.awt.Toolkit.getDefaultToolkit();
			} catch (Error e) {
				// log headless mode
				useGui = false;
			}
		}
		// start the pki server
		start();
		// wait until someone exits the programm
		String m = "Wait here";
		synchronized (m) {
			try {
				m.wait();
			} catch (Exception ignore) {
			}
		}
	}

	/**
	 * Initializes the network server, certification and database objects.
	 */
	public static void start() {
		System.out.println("Main.start called.");
		if (useGui && (panel == null)) {
			panel = new PKIPanel(settings);
			if (!panel.start()) {
				// must show a waiting message in the beginning
				System.out.println("ERR: Panel failed.");
				System.exit(-1);
			}
		}

		if (serv == null)
			serv = new PKIServer(settings);
		if (cert == null)
			cert = new PKICert(settings);
		if (data == null)
			data = new PKIDatabase(settings);

		// Enable debug..
		//      PKIServer.debug=true;
		//      PKIServer.PKIWorker.debug=true;
		//      PKICert.debug=true;
		//      PKIDatabase.debug=true;
		//      org.ssonet.net.SSONETProtocolStream.debug=true;
		//      org.ssonet.io.ByteIOStream.debug=true;

		if (startCert()) {
			if (startDatabase()) {
				if (startServer()) {
					if (useGui) {
						panel.setStarted();
					} else {
						System.out.println("Server is running.");
					}
				} else {
					if (!useGui) {
						System.out.println("ERR: Server failed.");
						System.exit(-1);
					}
				}
			} else {
				if (!useGui) {
					System.out.println("ERR: Database failed.");
					System.exit(-1);
				}
			}
		} else {
			if (!useGui) {
				System.out.println("ERR: Certifier failed.");
				System.exit(-1);
			}
		}
	}

	/**
	 * Starts the certification object.
	 */
	public static boolean startCert() {
		System.out.println("Main.startCert called.");
		// trys to start the certifier
		if (!cert.start()) {
			// notifies the gui about the result of starting
			if (useGui) {
				panel.setStopped(
					"<html>Certifier cannot be started.<p>"
						+ "Please check whether the keystore settings are correct,<p>"
						+ "and whether the keystore is initialized.</html>");
			}
			return false;
		}
		data.setCertifier(cert);
		serv.setCertifier(cert);
		if (useGui)
			panel.setCertifier(cert);
		return true;
	}

	/**
	 * Starts the database object.
	 */
	public static boolean startDatabase() {
		System.out.println("Main.startDatabase called.");
		// Close old connection if neccessary
		if (data.isConnected()) {
			data.stop();
			panel.setStopped("<html>Reconnecting to Database ...</html>");
		}
		// trys to start the database connection
		if (!data.start()) {
			// notifies the gui about the result of starting
			if (useGui) {
				panel.setStopped(
					"<html>Database connection cannot be started.<p>"
						+ "Please check whether the database settings are correct,<p>"
						+ "and whether the database is initialized.</html>");
			}
			return false;
		}
		cert.setDatabase(data);
		serv.setDatabase(data);
		if (useGui)
			panel.setDatabase(data);
		return true;
	}

	/**
	 * Starts the database object.
	 */
	public static boolean startServer() {
		System.out.println("Main.startServer called.");
		if (serv.isRunning()) {
			serv.stop();
			panel.setStopped("<html>Restarting the Server ...</html>");
			System.out.println("Restarting the Server ...");
		}
		// trys to starts the server
		if (!serv.start()) {
			if (useGui) {
				// notifies the GUI about the result of starting
				panel.setStopped(
					"<html>PKI Server cannot be started.<p>"
						+ "Please check whether the network and keystore settings are correct.</html>");
			}
			return false;
		}
		return true;
	}

	/**
	 * (Re-)initializes the database after settings change.
	 */
	public static void initDB() throws Exception {
		System.out.println("Main.initDB called.");
		// call the db initializing method
		data.stop();
		data.initDB();
		start();
	}

	/**
	 * Performs the -initdb commandline option.
	 */
	public static void initDBCommandline() {
		System.out.println("\nInitialize a Database for usage with the PKI");
		System.out.print(
			"\nData in this database will be deleted. Do you want to proceed? (y/n)");
		byte[] i = new byte[1];
		try {
			System.in.read(i);
			String c = new String(i);
			if (c.equals("Y")
				| c.equals("y")
				| c.equals("J")
				| c.equals("j")) {
				if (data == null)
					data = new PKIDatabase(settings);
				if (!data.initDB()) {
					throw new Exception("Initialization failed.");
				}
				System.out.println(
					"\nDatabase initialization finished successfully.");
			}
		} catch (Exception e) {
			System.out.println(
				"An error occured during database initialization: "
					+ e.getMessage());
			e.printStackTrace();
			System.out.println(
				"\nPlease check the database settings in data/PKISettings.properties");
		}
	}

	/**
	 * Stops and exits the program.
	 */
	public static void stop() {
		System.exit(0);
	}

}
