/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: PKICert.java,v 1.12 2003/12/16 08:36:58 tk4 Exp $
 */
package pki;

import java.io.FileInputStream;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.*;
import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Properties;

import org.bouncycastle.jce.X509Principal;
import org.bouncycastle.jce.X509V3CertificateGenerator;

/**
 * This class issues certificates.
 *
 * @author Thomas Kriegelstein
 * @version $Id: PKICert.java,v 1.12 2003/12/16 08:36:58 tk4 Exp $
 */
public class PKICert {

	public static boolean debug = false;

	Properties settings = null;
	PKIDatabase data = null;
	KeyStore ks = null;
	char[] carr = null;

	// Correct
	/** Registers the BouncyCastle security provider. */
	static {
		Security.addProvider(
			new org.bouncycastle.jce.provider.BouncyCastleProvider());
	}

	// Correct
	/**
	 * Stops the certificate object.
	 */
	public void stop() {
		data = null;
	}

	// Correct
	/**
	 * Initializes this certification object.
	 * @param settings the PKI settings
	 */
	public PKICert(Properties settings) {
		this.settings = settings;
	}

	// Correct
	/**
	 * Starts the certification object by loading the keystore.
	 * @return true, if starting was successful, false otherwise
	 */
	public boolean start() {
		carr =
			(settings.getProperty(Main.keyStorePassword) == null
				? null
				: settings.getProperty(Main.keyStorePassword).toCharArray());

		InputStream is = null;
		try {
			is =
				new FileInputStream(
					settings.getProperty(pki.Main.keyStoreFileName));
		} catch (Exception e) {
			is =
				getClass().getResourceAsStream(
					"/" + settings.getProperty(Main.keyStoreFileName));
		}
		try {
			ks = KeyStore.getInstance("JKS");
			ks.load(is, carr);
			return true;
		} catch (Exception e) {
			System.out.println("ERR: failed to load Keystore.");
			e.printStackTrace();
		}
		return false;
	}

	// Correct
	/**
	 * Registers the database object.
	 * @param data the database object
	 */
	public void setDatabase(PKIDatabase data) {
		this.data = data;
	}

	// Tested, via PKITest
	/**
	 * Checks a certificate.
	 * @param cert the certificate to check
	 * @return true, if the certificate is valid, false otherwise
	 */
	public boolean testCertificate(X509Certificate cert) {
		boolean result = false;
		try {
			if (debug)
				System.out.println("Testing: " + cert);
			X509Certificate pki = (X509Certificate)ks.getCertificate("pki");
			if (debug)
				System.out.println("Found: " + pki);
			PublicKey pk = pki.getPublicKey();
			if (debug)
				System.out.println("Checking validity");
			cert.checkValidity();
			if (debug)
				System.out.println("Checking signature");
			cert.verify(pk);
			if (debug)
				System.out.println("Passed. Certificate is valid.");
			result =
				(data.getPersonCertificate(cert.getSerialNumber()) != null);
			if (debug)
				System.out.println("Is known person: " + result);
		} catch (Exception e) {
			if (debug)
				System.out.println("FAILED!!! Certificate is invalid.");
			result = false;
		}
		return result;
	}

	// Tested, via PKITest
	/**
	 * Creates a person certificate and stores it into the database.
	 * @param pk the public key to certify
	 * @param p the properties of the certificate (the elements of a X.509 Distinguished Name)
	 * @param keypasswd the password for the private keystore
	 * @return the generated X.509 certificate
	 */
	public X509Certificate createPersonCertificate(
		PublicKey pk,
		Properties p,
		char[] keypasswd) {
		X509Certificate result = null;
		BigInteger serial = data.getNextSerialNumber();
		X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
		String subject = "CN=" + p.getProperty("CN");
		String[] arr = { "OU", "O", "L", "ST", "C" };
		for (int i = 0; i < arr.length; i++) {
			String t = p.getProperty(arr[i]);
			if (t != null && t.length() > 0)
				subject += "," + arr[i] + "=" + t;
		} // Alles eingesammelt
		certGen.setSubjectDN(new X509Principal(subject));
		GregorianCalendar gregorianCalendar =
			(GregorianCalendar)Calendar.getInstance();
		certGen.setNotBefore(gregorianCalendar.getTime());
		gregorianCalendar.add(Calendar.YEAR, 1);
		certGen.setNotAfter(gregorianCalendar.getTime());
		certGen.setPublicKey(pk);
		try {
			X509Certificate cert = (X509Certificate)ks.getCertificate("pki");
			certGen.setSignatureAlgorithm(cert.getSigAlgName());
			certGen.setIssuerDN(
				new X509Principal(cert.getSubjectDN().getName()));
			certGen.setSerialNumber(serial);

			PrivateKey secKey = (PrivateKey)ks.getKey("pki", keypasswd);
			result = certGen.generateX509Certificate(secKey);
			secKey = null;
		} catch (Exception e) {
			System.out.println("\n FAILURE: " + e.getMessage() + "\n");
			e.printStackTrace();
			return null;
		}
		// create it...
		data.addPersonCertificate(result);
		return result;
	}

	// Correct
	/**
	 * Creates a certificate relative to the person certificate with the
	 * given serial number.
	 * @param serial the serial number of the person certificate
	 * @param pk the public key to certify
	 * @param keypasswd the password of the private keystore
	 * @return the generated X.509 certificate
	 */
	public X509Certificate createCertificate(
		BigInteger serial,
		PublicKey pk,
		char[] keypasswd) {
		X509Certificate result = null;
		X509Certificate tmp = data.getPersonCertificate(serial);
		result = createCertificate(tmp, pk, keypasswd);
		return result;
	}

	// Tested via PKITest
	/**
	 * Creates a certificate relative to a given person certificate.
	 * @param person the person certifiate
	 * @param pk the public key to certify
	 * @param keypasswd the password for the private key store
	 * @return the generated X.509 certificate
	 */
	public X509Certificate createCertificate(
		X509Certificate person,
		PublicKey pk,
		char[] keypasswd) {
		BigInteger serial = person.getSerialNumber();
		BigInteger next = data.getNextSerialNumber();
		X509Certificate result = null;

		X509V3CertificateGenerator certGen = new X509V3CertificateGenerator();
		String subject = "CN=" + next.toString();
		certGen.setSubjectDN(new X509Principal(subject));
		GregorianCalendar gregorianCalendar =
			(GregorianCalendar)Calendar.getInstance();
		certGen.setNotBefore(gregorianCalendar.getTime());
		gregorianCalendar.add(Calendar.YEAR, 1);
		certGen.setNotAfter(gregorianCalendar.getTime());
		certGen.setPublicKey(pk);
		try {
			X509Certificate cert = (X509Certificate)ks.getCertificate("pki");
			certGen.setSignatureAlgorithm(cert.getSigAlgName());
			certGen.setIssuerDN(
				new X509Principal(cert.getSubjectDN().getName()));
			certGen.setSerialNumber(next);

			PrivateKey secKey = (PrivateKey)ks.getKey("pki", keypasswd);
			result = certGen.generateX509Certificate(secKey);
			secKey = null;
		} catch (Exception e) {
			System.out.println("\n FAILURE: " + e.getMessage() + "\n");
			e.printStackTrace();
			return null;
		}
		data.addCertificate(serial, result);
		return result;
	}
}
