/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: PKIDatabase.java,v 1.14 2003/12/16 08:36:58 tk4 Exp $
 */
package pki;

import java.io.*;
import java.math.BigInteger;
import java.security.NoSuchProviderException;
import java.security.cert.*;
import java.sql.*;
import java.util.Hashtable;
import java.util.Properties;

/**
 * This class manages the PKIs database.
 *
 * @author Thomas Kriegelstein
 * @version $Id: PKIDatabase.java,v 1.14 2003/12/16 08:36:58 tk4 Exp $
 */
public class PKIDatabase implements Runnable {

	public static boolean debug = false;

	// variables used for the file database
	private Hashtable persons = null;
	private Hashtable roles = null;
	private Hashtable pseudos = null;
	private BigInteger last = null;
	private String connectedDBFileName = null;

	private boolean isConnected = false;
	private Thread t = null;

	// variables used for the mysql database
	/** The database connection object.              */
	protected Connection c = null;

	private PKICert cert = null;
	private Properties settings = null;

	private int dbType;
	public static final int DB_FILE = 0;
	public static final int DB_MYSQL = 1;

	/**
	 * Constructor for selecting the database type to use
	 * @param settings Settings for Database
	 */
	public PKIDatabase(Properties settings) {
		this.settings = settings;
	}

	public boolean isConnected() {
		return isConnected;
	}

	/**
	 * Returns the next free serial number, and stores it into database
	 */
	public BigInteger getNextSerialNumber() {
		switch (dbType) {
			case DB_FILE :
				last = last.add(BigInteger.ONE);
				return last;
			case DB_MYSQL :
				try {
					// Lock table config
					Statement s = c.createStatement();
					s.execute("LOCK TABLES config WRITE");
					// get last serial number
					s =
						c.createStatement(
							ResultSet.TYPE_SCROLL_INSENSITIVE,
							ResultSet.CONCUR_READ_ONLY);
					ResultSet r =
						s.executeQuery(
							"SELECT value FROM config WHERE name='lastSerial'");
					r.first();
					last = new BigInteger(r.getString(1));
					// increment the number
					last = last.add(BigInteger.ONE);
					// save new number in database
					PreparedStatement sp =
						c.prepareStatement(
							"UPDATE config SET value=? WHERE name='lastSerial'");
					sp.setString(1, last.toString());
					sp.executeUpdate();
					sp.close();
					// release lock
					s.execute("UNLOCK TABLES");
					s.close();
				} catch (SQLException sqle) {
					sqle.printStackTrace();
					System.err.println(
						"Unable to increment serial number. Stop.");
					System.exit(-1);
				}
				return last;
			default :
				return null;
		}
	}

	/**
	 * Returns last (used) serial number
	 */
	public BigInteger getLastSerial() {
		switch (dbType) {
			case DB_FILE :
				return last;
			case DB_MYSQL :
				try {
					Statement s = c.createStatement();
					// get last serial number
					s =
						c.createStatement(
							ResultSet.TYPE_SCROLL_INSENSITIVE,
							ResultSet.CONCUR_READ_ONLY);
					ResultSet r =
						s.executeQuery(
							"SELECT value FROM config WHERE name='lastSerial'");
					r.first();
					last = new BigInteger(r.getString(1));
					s.close();
				} catch (SQLException sqle) {
					sqle.printStackTrace();
					System.err.println(
						"Unable to get last serial number. Stop.");
					System.exit(-1);
				}
				return last;
			default :
				return null;
		}
	}

	/**
	 * Returns a certificate for a person pseudonym according to the given serial
	 * number
	 */
	public X509Certificate getPersonCertificate(BigInteger serial) {
		if (debug)
			System.out.println("DBG: getPersonCertificate(" + serial + ")");
		X509Certificate result = null;
		switch (dbType) {
			case DB_FILE :
				if (serial.compareTo(this.last) <= 1) {
					result = (X509Certificate)persons.get(serial);
				} else {
					// Fehlerbehandlung ?!
				}
				break;
			case DB_MYSQL :
				try {
					PreparedStatement s =
						c.prepareStatement(
							"SELECT cert FROM persons WHERE serial = ?",
							ResultSet.TYPE_SCROLL_INSENSITIVE,
							ResultSet.CONCUR_READ_ONLY);
					s.setBytes(1, serial.toByteArray());
					ResultSet r = s.executeQuery();
					byte[] ser_cert = null;
					r.beforeFirst();
					if (r.next()) {
						ser_cert = r.getBytes(1);
						// deserialize certificate
						try {
							result =
								(X509Certificate)CertificateFactory
									.getInstance("X.509", "BC")
									.generateCertificate(
										new ByteArrayInputStream(ser_cert));
						} catch (CertificateException ce) {
							// Exception is thrown on parsing errors
							result = null;
						} catch (NoSuchProviderException nspe) {
							System.out.println(
								"BounyCastle Provider not found. Stop.");
							nspe.printStackTrace();
							System.exit(-1);
						}
					}
					r.close();
					s.close();
				} catch (SQLException sqle) {
					System.err.println(
						"Error in getPersonCertificate() from database. Stop.");
					sqle.printStackTrace();
					System.exit(-1);
				}
		}
		return result;
	}

	/**
	 * Returns a certificate for a role pseudonym according to the given serial
	 * number
	 */
	public X509Certificate getRoleCertificate(BigInteger serial) {
		X509Certificate result = null;
		switch (dbType) {
			case DB_FILE :
				if (serial.compareTo(this.last) <= 1) {
					result = (X509Certificate)roles.get(serial);
				} else {
					// Fehlerbehandlung ?!
				}
				break;
			case DB_MYSQL :
				try {
					PreparedStatement s =
						c.prepareStatement(
							"SELECT cert FROM roles WHERE serial = ?",
							ResultSet.TYPE_SCROLL_INSENSITIVE,
							ResultSet.CONCUR_READ_ONLY);
					s.setBytes(1, serial.toByteArray());
					ResultSet r = s.executeQuery();
					byte[] ser_cert = null;
					r.beforeFirst();
					if (r.next()) {
						ser_cert = r.getBytes(1);
						// deserialize certificate
						try {
							result =
								(X509Certificate)CertificateFactory
									.getInstance("X.509", "BC")
									.generateCertificate(
										new ByteArrayInputStream(ser_cert));
						} catch (CertificateException ce) {
							// Exception is thrown on parsing errors
							result = null;
						} catch (NoSuchProviderException nspe) {
							System.out.println(
								"BounyCastle Provider not found. Stop.");
							nspe.printStackTrace();
							System.exit(-1);
						}
					}
					r.close();
					s.close();
				} catch (SQLException sqle) {
					System.err.println(
						"Error in getPersonCertificate() from database. Stop.");
					sqle.printStackTrace();
					System.exit(-1);
				}
		}
		return result;
	}

	/**
	 * Returns a certificate for a pseudonym according to the given serial
	 * number
	 */
	public X509Certificate getCertificate(BigInteger serial) {
		if (debug)
			System.out.println("DBG: getCertificate(" + serial + ")");
		X509Certificate result = null;
		switch (dbType) {
			case DB_FILE :
				if (serial.compareTo(this.last) <= 1) {
					result = (X509Certificate)pseudos.get(serial);
				} else {
					// Fehlerbehandlung ?!
				}
				break;
			case DB_MYSQL :
				try {
					{
						X509Certificate c = getPersonCertificate(serial);
						if (c != null)
							return c;
					}
					PreparedStatement s =
						c.prepareStatement(
							"SELECT cert, person FROM pseudos",
							ResultSet.TYPE_SCROLL_INSENSITIVE,
							ResultSet.CONCUR_READ_ONLY);
					ResultSet r = s.executeQuery();
					byte[] ser_cert = null;
					r.beforeFirst();
					while (r.next()) {
						ser_cert = r.getBytes(1);
						byte[] p = r.getBytes(2);
						// deserialize certificate
						try {
							result =
								(X509Certificate)CertificateFactory
									.getInstance("X.509", "BC")
									.generateCertificate(
										new ByteArrayInputStream(ser_cert));
							if (result.getSerialNumber().compareTo(serial)
								== 0) {
								if (debug)
									System.out.println(
										"DBG: found certificate for " + serial);
								if (debug)
									System.out.println(
										"DBG: serial number is "
											+ new BigInteger(p));
								result =
									getPersonCertificate(new BigInteger(p));
								if (result != null)
									break;
								else {
									if (debug)
										System.out.println(
											"DBG: person certificate missing!");
								}
							} else {
								result = null;
								continue;
							}
						} catch (CertificateException ce) {
							// Exception is thrown on parsing errors
							result = null;
						} catch (NoSuchProviderException nspe) {
							System.out.println(
								"BounyCastle Provider not found. Stop.");
							nspe.printStackTrace();
							System.exit(-1);
						}
					}
					r.close();
					s.close();
				} catch (SQLException sqle) {
					System.err.println(
						"Error in getPersonCertificate() from database. Stop.");
					sqle.printStackTrace();
					System.exit(-1);
				}
		}
		return result;
	}

	/**
	 * Inserts a role certificate into the database
	 */
	public void addRoleCertificate(BigInteger person, X509Certificate cert) {
		switch (dbType) {
			case DB_FILE :
				roles.put(cert, person);
				break;
			case DB_MYSQL :
				try {
					PreparedStatement s =
						c.prepareStatement(
							"INSERT INTO roles VALUES ( ? , ? )");
					s.setBytes(1, cert.getEncoded());
					s.setBytes(2, person.toByteArray());
					s.executeUpdate();
					s.close();
				} catch (SQLException sqle) {
					System.err.println(
						"Error in addRoleCertificate() into database. Stop.");
					sqle.printStackTrace();
					System.exit(-1);
				} catch (CertificateEncodingException cee) {
					System.err.println(
						"Unable to encode Certifocate. Implementation Error. Stop.");
					cee.printStackTrace();
					System.exit(-1);
				}
		}
	}

	/**
	 * Inserts a person certificate into the database
	 */
	public void addPersonCertificate(X509Certificate cert) {
		switch (dbType) {
			case DB_FILE :
				persons.put(cert.getSerialNumber(), cert);
				break;
			case DB_MYSQL :
				try {
					PreparedStatement s =
						c.prepareStatement(
							"INSERT INTO persons VALUES ( ? , ? )");
					s.setBytes(1, cert.getSerialNumber().toByteArray());
					s.setBytes(2, cert.getEncoded());
					s.executeUpdate();
					s.close();
				} catch (SQLException sqle) {
					System.err.println(
						"Error in addRoleCertificate() into database. Stop.");
					sqle.printStackTrace();
					System.exit(-1);
				} catch (CertificateEncodingException cee) {
					System.err.println(
						"Unable to encode Certifocate. Implementation Error. Stop.");
					cee.printStackTrace();
					System.exit(-1);
				}
		}
	}

	/**
	 * Inserts a pseudonym certificate into the database
	 */
	public void addCertificate(BigInteger person, X509Certificate cert) {
		switch (dbType) {
			case DB_FILE :
				pseudos.put(cert, person);
				break;
			case DB_MYSQL :
				try {
					PreparedStatement s =
						c.prepareStatement(
							"INSERT INTO pseudos VALUES ( ? , ? )");
					s.setBytes(1, cert.getEncoded());
					s.setBytes(2, person.toByteArray());
					s.executeUpdate();
					s.close();
				} catch (SQLException sqle) {
					System.err.println(
						"Error in addRoleCertificate() into database. Stop.");
					sqle.printStackTrace();
					System.exit(-1);
				} catch (CertificateEncodingException cee) {
					System.err.println(
						"Unable to encode Certificate. Implementation Error. Stop.");
					cee.printStackTrace();
					System.exit(-1);
				}
		}
	}

	/**
	 * Sets the certifier object to be used by this database instance.
	 */
	public void setCertifier(PKICert cert) {
		this.cert = cert;
	}

	/**
	 * Opens the database.
	 */
	public boolean start() {
		// Start database thread
		if (t == null) {
			// ensure, that only one single database thread is started
			t = new Thread(this);
			t.setName("pki.PKIDatabase.ShutdownHook");
			Runtime.getRuntime().addShutdownHook(t);
		}
		// try to open the database
		if (settings.getProperty(Main.dbType).equals("file")) {
			dbType = DB_FILE;
		} else {
			dbType = DB_MYSQL;
		}
		switch (dbType) {
			case DB_FILE :
				return openDBFile();
			case DB_MYSQL :
				return openDBMysql();
			default :
				return false;
		}
	}

	/**
	 * Initializes a new database.
	 */
	public boolean initDB() {
		if (settings.getProperty(Main.dbType).equals("file")) {
			dbType = DB_FILE;
		} else {
			dbType = DB_MYSQL;
		}
		switch (dbType) {
			case DB_FILE :
				return initDBFile();
			case DB_MYSQL :
				return initDBMysql();
			default :
				return false;
		}
	}

	/**
	 * Opens a file database.
	 */
	public boolean openDBFile() {
		// More magic, encryption etc. May use PKICert for that
		System.out.println("openDBFile called");
		boolean result = false;
		try {
			FileInputStream fis =
				new FileInputStream(settings.getProperty(Main.fileDBFileName));
			connectedDBFileName = settings.getProperty(Main.fileDBFileName);
			// DecryptInputStream ...
			// VerifyInputStream ...
			ObjectInputStream ois = new ObjectInputStream(fis);
			BigInteger l = (BigInteger)ois.readObject();
			Hashtable e = (Hashtable)ois.readObject();
			Hashtable r = (Hashtable)ois.readObject();
			Hashtable p = (Hashtable)ois.readObject();
			last = l;
			persons = e;
			roles = r;
			pseudos = p;
			isConnected = true;
			result = true;
		} catch (Exception e) {
			System.err.println(
				"Error opening file DB: "
					+ settings.getProperty(Main.fileDBFileName));
		}
		return result;
	}

	/**
	 * Initializes the file database.
	 */
	public boolean initDBFile() {
		System.out.println("INF: Reinitialize File database");
		// More magic, encryption etc. May use PKICert for that
		last = new BigInteger("0");
		persons = new Hashtable();
		roles = new Hashtable();
		pseudos = new Hashtable();
		isConnected = true;
		save();
		return true;
	}

	/**
	 * Opens a MySQL database.
	 */
	public boolean openDBMysql() {
		boolean result = false;
		try {
			// initialize db driver
			java.lang.Class.forName("org.gjt.mm.mysql.Driver").newInstance();
			// connect to the database
			c =
				DriverManager.getConnection(
					"jdbc:mysql://"
						+ settings.getProperty(Main.mysqlHost)
						+ (settings.getProperty(Main.mysqlPort).equals("")
							? ""
							: (":" + settings.getProperty(Main.mysqlPort)))
						+ "/"
						+ settings.getProperty(Main.mysqlDBName),
					settings.getProperty(Main.mysqlUsername),
					settings.getProperty(Main.mysqlPassword));
			try {
				// check, whether this database is initialized
				Statement s =
					c.createStatement(
						ResultSet.TYPE_SCROLL_INSENSITIVE,
						ResultSet.CONCUR_READ_ONLY);
				ResultSet r = s.executeQuery("SELECT serial FROM persons");
				r = s.executeQuery("SELECT person FROM roles");
				r = s.executeQuery("SELECT person FROM pseudos");
				r = s.executeQuery("SELECT name FROM config");
				r =
					s.executeQuery(
						"SELECT value FROM config WHERE name='lastSerial'");
				if (r.first() == false) {
					throw new Exception("Last serial is not initialized.");
				}
				r.close();
				s.close();
				isConnected = true;
				result = true;
			} catch (Exception e) {
				System.err.println(
					"WAR: Could not find initialized mysql database: "
						+ "jdbc:mysql://"
						+ settings.getProperty(Main.mysqlHost)
						+ (settings.getProperty(Main.mysqlPort).equals("")
							? ""
							: (":" + settings.getProperty(Main.mysqlPort)))
						+ "/"
						+ settings.getProperty(Main.mysqlDBName));
				result = false;
			}
			if (result == true) {
				last = BigInteger.ZERO;
			}
		} catch (Exception e) {
			System.err.println(
				"WAR: Could not connect to mysql database: "
					+ "jdbc:mysql://"
					+ settings.getProperty(Main.mysqlHost)
					+ (settings.getProperty(Main.mysqlPort).equals("")
						? ""
						: (":" + settings.getProperty(Main.mysqlPort)))
					+ "/"
					+ settings.getProperty(Main.mysqlDBName));
			e.printStackTrace();
			result = false;
		}
		return result;
	}

	/**
	 * Initializes a MySQL database.
	 */
	public boolean initDBMysql() {
		boolean result = false;
		System.out.println("INF: (Re)initialize MySQL database");
		try {
			// initialize db driver
			java.lang.Class.forName("org.gjt.mm.mysql.Driver").newInstance();
			// erase the content of the old database
			if ((c == null) || c.isClosed()) {
				c =
					DriverManager.getConnection(
						"jdbc:mysql://"
							+ settings.getProperty(Main.mysqlHost)
							+ (settings.getProperty(Main.mysqlPort).equals("")
								? ""
								: (":" + settings.getProperty(Main.mysqlPort)))
							+ "/"
							+ settings.getProperty(Main.mysqlDBName),
						settings.getProperty(Main.mysqlUsername),
						settings.getProperty(Main.mysqlPassword));
			}
			Statement s = c.createStatement();
			s.executeUpdate(
				"DROP TABLE IF EXISTS persons, roles, pseudos, config");
			// try to initialize a new database
			// create tables
			s.executeUpdate(
				"CREATE TABLE config ("
					+ "PRIMARY KEY (name), "
					+ "name VARCHAR(200) NOT NULL, "
					+ "value VARCHAR(200) )");
			s.executeUpdate(
				"CREATE TABLE persons ("
					+ "PRIMARY KEY (serial), "
					+ "serial BINARY(40) NOT NULL, "
					+ "cert BLOB NOT NULL)");
			s.executeUpdate(
				"CREATE TABLE roles ("
					+ "cert BLOB NOT NULL, "
					+ "person BINARY(40) NOT NULL)");
			s.executeUpdate(
				"CREATE TABLE pseudos ("
					+ "cert BLOB NOT NULL ,"
					+ "person BINARY(40) NOT NULL)");
			s.executeUpdate(
				"INSERT INTO config (name, value) VALUES ('lastSerial', '0')");
			s.close();
			result = true;
		} catch (SQLException sqle) {
			System.err.println(
				"Unable to initialize new database. Please adjust settings.");
			sqle.printStackTrace();
			result = false;
		} catch (Exception e) {
			System.err.println("Database driver not found.");
		}
		return result;
	}

	/**
	 * Saves the database and closes the connection to it.
	 */
	private void save() {
		switch (dbType) {
			case DB_FILE :
				if (isConnected) {
					BigInteger l = last;
					Hashtable e = persons;
					Hashtable r = roles;
					Hashtable p = pseudos;
					try {
						int i = connectedDBFileName.lastIndexOf("/");
						try {
							new File(connectedDBFileName.substring(0, i))
								.mkdir();
						} catch (Exception ignore) {
						}
						FileOutputStream fos =
							new FileOutputStream(connectedDBFileName);
						// EncryptOutputStream ...
						// SignOutputStream ...
						ObjectOutputStream oos = new ObjectOutputStream(fos);
						oos.writeObject(l);
						oos.writeObject(e);
						oos.writeObject(r);
						oos.writeObject(p);
						oos.close();
					} catch (Exception ex) {
						System.err.println("ERR: PKIDatabase save failed.");
						ex.printStackTrace();
					}
				}
				break;
			case DB_MYSQL :
				try {
					// close the connection
					c.close();
				} catch (SQLException sqle) {
					sqle.printStackTrace();
					System.err.println(
						"Unable to close connection to database. Don't care.");
				}
		}
		isConnected = false;
	}

	/**
	 * Stops the database object.
	 */
	public void stop() {
		save();
		cert = null;
	}

	public void run() {
		if (cert != null) {
			System.out.println("INF: Database shutting down.");
			save();
		}
	}
}
