/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: PKIServer.java,v 1.18 2004/02/12 08:13:00 tk4 Exp $
 */
package pki;

import java.io.*;
import java.math.BigInteger;
import java.net.SocketException;
import java.security.KeyStore;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Calendar;
import java.util.GregorianCalendar;
import java.util.Properties;

import org.bouncycastle.jce.X509Principal;
import org.bouncycastle.jce.X509V3CertificateGenerator;
import org.ssonet.net.*;
import org.ssonet.net.impl.*;

/**
 * Network server for online issuing of certified pseudonyms.
 *
 * @author Thomas Kriegelstein
 * @version $Id: PKIServer.java,v 1.18 2004/02/12 08:13:00 tk4 Exp $
 */
public class PKIServer
	implements Runnable, SSONETContextFactory, CertificateListener {

	public static boolean debug = false;

	private Properties settings = null;
	private boolean isRunning = false;
	private Thread serverThread = null;
	SSONETServerSocket ss = null;
	BaseConfiguration conf = new BaseConfiguration();

	protected X509Certificate sigCert = null;
	protected X509Certificate cipCert = null;
	protected PrivateKey sigKey = null;
	protected PrivateKey cipKey = null;

	protected PKICert certifier = null;
	protected PKIDatabase database = null;

	public boolean isRunning() {
		return isRunning;
	}

	public void setDatabase(PKIDatabase data) {
		database = data;
	}
	public void setCertifier(PKICert cert) {
		certifier = cert;
	}

	public PKIServer(Properties settings) {
		this.settings = settings;
	}

	public synchronized boolean start() {
		if (debug)
			System.out.println("PKIServer.start called.");
		boolean result = true;
		if (database != null && certifier != null) {
			InputStream is = null;
			try {
				is =
					new FileInputStream(
						settings.getProperty(Main.keyStoreFileName));
			} catch (Exception e) {
				is =
					getClass().getResourceAsStream(
						"/" + settings.getProperty(Main.keyStoreFileName));
			}
			try {
				char[] storepass =
					settings.getProperty(Main.keyStorePassword).toCharArray();
				char[] key1pass =
					settings.getProperty(Main.secretKeyPassword).toCharArray();
				char[] key2pass =
					settings.getProperty(Main.secretKeyPassword).toCharArray();
				KeyStore ks = KeyStore.getInstance("JKS");
				ks.load(is, storepass);
				sigCert = (X509Certificate)ks.getCertificate("pki");
				sigKey = (PrivateKey)ks.getKey("pki", key1pass);
				cipCert = (X509Certificate)ks.getCertificate("cipher");
				cipKey = (PrivateKey)ks.getKey("cipher", key2pass);

				X509V3CertificateGenerator certGen = null;
				GregorianCalendar gregorianCalendar = null;

				certGen = new X509V3CertificateGenerator();
				certGen.setSubjectDN(new X509Principal("CN=Cipherkey"));
				gregorianCalendar = (GregorianCalendar)Calendar.getInstance();
				certGen.setNotBefore(gregorianCalendar.getTime());
				gregorianCalendar.add(Calendar.YEAR, 1);
				certGen.setNotAfter(gregorianCalendar.getTime());
				certGen.setPublicKey(cipCert.getPublicKey());
				certGen.setSignatureAlgorithm("SHA1withDSA");
				certGen.setIssuerDN(new X509Principal("CN=RootCA (Test!)"));
				certGen.setSerialNumber(BigInteger.ONE);
				cipCert = certGen.generateX509Certificate(sigKey);
				// Create ServerSocket and start network thread
				ss = new SSONETServerSocket(7777, this);
				isRunning = true;
				serverThread = new Thread(this);
				serverThread.start();
				if (debug)
					System.out.println("PKIServer: Init completed.");
				result = true;
			} catch (Exception e) {
				result = false;
				e.printStackTrace();
			}
		} else {
			result = false;
		}
		return result;
	}

	public int checkCertificate(SSONETContext context, X509Certificate cert) {
		return certifier.testCertificate(cert) ? VALID : INVALID;
	}
	public SSONETContext createContext() {
		SSONETContext result = new SSONETContext();
		try {
			result.setBaseConfiguration(conf);
		} catch (Exception e) {
			e.printStackTrace();
		}
		//result.addActionIDListener(this);
		result.addCertificateListener(this);

		result.setOwnTestCertificate(sigCert);
		result.setOwnCipherCertificate(cipCert);
		result.setOwnSigKey(sigKey);
		result.setOwnCiphKey(cipKey);
		ActionConfiguration ac = new ActionConfiguration(3, 3, 3, 3);
		result.setOwnActionConfiguration(ac);
		if (debug)
			System.out.println("PKIServer: Context created.");
		return result;
	}

	public void run() {
		try {
			ss.setSoTimeout(10000);
		} catch (SocketException se) {
			System.out.println("Cannot set Timeout on ServerSocket. Stop.");
			System.exit(-1);
		}
		while (isRunning) {
			// Endlosschleife
			try {
				if (debug)
					System.out.println(
						"PKIServer: Wait for connection on: " + ss);
				SSONETServerClientSocket cs =
					(SSONETServerClientSocket)ss.accept();
				// prevent clientSocket from inheriting timeout
				cs.setSoTimeout(0);
				if (debug)
					System.out.println("PKIServer: Got connection: " + cs);
				//   Verbindung annehmen
				PKIWorker worker =
					new PKIWorker(
						certifier,
						database,
						settings
							.getProperty(Main.secretKeyPassword)
							.toCharArray());
				//   PKIWorker fuer Verbindung starten
				worker.set(cs);
				worker.start();
			} catch (IOException ioe) {
				if (ioe instanceof InterruptedIOException) {
					// ignore
				} else if (isRunning == false) {
					// Do nothing. Thread has been stopped
				} else {
					ioe.printStackTrace();
					isRunning = false;
				}
			}
			// Schleifenende
		}
		System.out.println("ServerThread stops.");
	}

	public synchronized void stop() {
		if (isRunning) {
			isRunning = false;
			try {
				ss.close();
				System.out.println("Closed old ServerSocket.");
				synchronized (this) {
					notifyAll();
				}
				serverThread.join();
				System.out.println("Old ServerThread died.");
				serverThread = null;
			} catch (Exception e) {
				e.printStackTrace();
				System.out.println("Error closing the ServerSocket.");
			};
		}
	}

	public final static class PKIWorker
		extends Thread
		implements ConnectionListener {

		public static boolean debug = false;

		PKICert certifier = null;
		PKIDatabase database = null;

		char[] keypw = null;

		SSONETServerClientSocket todo = null;

		public PKIWorker(
			PKICert certifier,
			PKIDatabase database,
			char[] keypw) {
			this.certifier = certifier;
			this.database = database;
			this.keypw = keypw;
		}

		public synchronized void connectionStateChanged(
			SSONETContext context,
			int currentState) {
			if (currentState == RUNNING_STATE)
				notify();
		}

		public synchronized void waitConnect(SSONETContext context) {
			context.addConnectionListener(this);
			while (context.getPartnerCipherCertificate() == null) {
				try {
					wait();
				} catch (InterruptedException ie) {
				}
			}
			context.removeConnectionListener(this);
		}

		public void run() {
			if (debug)
				System.out.println("PKIWorker: working on: " + todo);
			InputStream is = null;
			OutputStream os = null;
			SSONETContext context = todo.getContext();
			try {
				// 1. handshake
				if (debug)
					System.out.println("PKIWorker: wait for handshake...");
				this.waitConnect(context);
				if (debug)
					System.out.println("PKIWorker: ... handshake done.");
				is = todo.getInputStream();
				os = todo.getOutputStream();
				// 2. cert now available, fetch and test
				X509Certificate cert = context.getPartnerTestCertificate();
				// eigentlich schon im Certificatelistener... sicher ist sicher
				if (!(certifier.testCertificate(cert)
					&& database.getPersonCertificate(cert.getSerialNumber())
						!= null)) {
					if (debug)
						System.out.println("PKIWorker: illegal cert (BAD!!!!)");
					if (debug)
						System.out.println(
							"PKIWorker: reason: (cert==null)   "
								+ (cert == null));
					if (debug)
						System.out.println(
							"PKIWorker: reason: (test failed)  "
								+ (!certifier.testCertificate(cert)));
					if (debug)
						System.out.println(
							"PKIWorker: reason: (unknown cert) "
								+ (null
									== database.getPersonCertificate(
										cert.getSerialNumber())));
					try {
						todo.close();
					} catch (Exception ignore) {
						// log error
					}
				} else {
					if (debug)
						System.out.println("PKIWorker: valid cert (GOOD)");
					// read new certificate
					int len = new DataInputStream(is).readInt();
					byte[] arr = new byte[len];
					for (int j = 0; j < len - 1; j += is.read(arr, j, len - j));
					// encoded certificate wieder umwandeln in Objekt...
					CertificateFactory cf =
						CertificateFactory.getInstance("X.509");
					X509Certificate c =
						(X509Certificate)cf.generateCertificate(
							new ByteArrayInputStream(arr));
					// ...
					c.checkValidity();
					// sollte stimmen!
					PublicKey pk = c.getPublicKey();
					// certifikat unterschreiben
					// PKICert benutzen!!!
					X509Certificate nc =
						certifier.createCertificate(cert, pk, keypw);
					byte[] result = nc.getEncoded();
					new DataOutputStream(os).writeInt(result.length);
					os.write(result);
					os.flush();
					if (debug)
						System.out.println("PKIWorker: done.");
					os.close();
				}
			} catch (Exception e) {
				e.printStackTrace();
				try {
					os.close();
				} catch (Exception ignore) {
				}
			}
		} // run

		public void set(SSONETServerClientSocket socket) {
			todo = socket;
		}

		/* (non-Javadoc)
		 * @see org.ssonet.net.ConnectionListener#negotiationRequestReceived(org.ssonet.net.SSONETContext, int)
		 */
		public void negotiationRequestReceived(SSONETContext context, int negotiationType) {
		}

		/* (non-Javadoc)
		 * @see org.ssonet.net.ConnectionListener#decideSecurityGoal(int, org.ssonet.net.ActionConfiguration, org.ssonet.net.SSONETContext)
		 */
		public boolean decideSecurityGoal(int securityGoal, ActionConfiguration ownConfiguration, SSONETContext context) {
			return false;
		} // set

	} // PKIWorker

} // PKIServer
