/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: Query.java,v 1.6 2003/12/16 08:36:58 tk4 Exp $
 */
package pki;

import java.io.File;
import java.io.FileInputStream;
import java.math.BigInteger;
import java.security.cert.X509Certificate;
import java.util.Properties;

/**
 * The query class does command line pseudonymity revocation.
 *
 * @author Thomas Kriegelstein
 * @version $Id: Query.java,v 1.6 2003/12/16 08:36:58 tk4 Exp $
 */
public class Query {
	/** settings for configuration of the PKI */
	private static Properties settings = new Properties();
	public static final String keyStoreFileName = "keyStoreFileName";
	public static final String secretKeyPassword = "secretKeyPassword";
	public static final String keyStorePassword = "keyStorePassword";
	public static final String dbType = "dbType";
	public static final String fileDBFileName = "fileDBFileName";
	public static final String mysqlHost = "mysqlHost";
	public static final String mysqlPort = "mysqlPort";
	public static final String mysqlDBName = "mysqlDBName";
	public static final String mysqlUsername = "mysqlUsername";
	public static final String mysqlPassword = "mysqlPassword";
	private static String settingsFileName = "data/pkiSettings.properties";

	static PKICert cert = null;
	static PKIDatabase data = null;

	/**
	 * Prints a usage message to standard out
	 */
	private static void usage() {
		System.out.println("Usage: pki.Query [options] command");
		String[] opts =
			{
				"Options are:",
				"  -keypass      %s  use secret for server operation",
				"  -storephrase  %s  use secret to validate the integrity of the keystore",
				"  -mysql            use mysql database instead of file database",
				"Commands are:",
				"  -query        %s  display personcertificate for serial number",
				"  -help             print this help message and exit" };
		for (int i = 0; i < opts.length; i++)
			System.out.println(opts[i]);
	}

	/**
	 * Reads settings from file
	 */
	private static void loadSettings() {
		try {
			File sf = new File(settingsFileName);
			settings.load(new FileInputStream(sf));
		} catch (Exception e) {
			System.out.println(
				"Unable to read Settings from file: " + settingsFileName);
		}
	}

	/**
	 * Sets default settings for the PKI Server
	 */
	private static void setDefaultSettings() {
		settings = new Properties();
		settings.setProperty(keyStoreFileName, "data/pkikeys");
		settings.setProperty(secretKeyPassword, "secret");
		settings.setProperty(keyStorePassword, "secret");
		settings.setProperty(dbType, "mysql");
		settings.setProperty(fileDBFileName, "data/pkidata");
		settings.setProperty(mysqlHost, "dud72");
		settings.setProperty(mysqlPort, "3306");
		settings.setProperty(mysqlDBName, "pki");
		settings.setProperty(mysqlUsername, "pkiadmin");
		settings.setProperty(mysqlPassword, "pkiadmin");
	}

	/**
	 * Main method. Handles commandline options.
	 */
	public static void main(String[] args) throws Exception {
		//System.out.println("Query.main called.");
		setDefaultSettings();
		// load settings from file
		loadSettings();
		String serial = null;
		// get command line options
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("?") || args[i].equals("-help")) {
				usage();
				System.exit(0);
				break;
			} else if (args[i].equals("-storephrase")) {
				i += 1;
				if (i < args.length) {
					settings.setProperty(keyStorePassword, args[i]);
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else if (args[i].equals("-keypass")) {
				i += 1;
				if (i < args.length) {
					settings.setProperty(secretKeyPassword, args[i]);
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else if (args[i].equals("-mysql")) {
				settings.setProperty(dbType, "mysql");
			} else if (args[i].equals("-query")) {
				i += 1;
				if (i < args.length) {
					serial = args[i];
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else {
				usage();
				System.exit(0);
				break;
			}
		}
		if (serial == null) {
			System.out.println("No command given.");
			usage();
			System.exit(-1);
		} else
			try {
				BigInteger b = null;
				try {
					b = new BigInteger(serial);
				} catch (Exception e) {
					b = null;
				}
				if (b != null) {
					System.out.println("INF: " + serial + " is a number.");
				} else {
					X509Certificate in = PKIUtils.getFileCertificate(serial);
					if (in == null) {
						System.out.println(
							"ERR: Could not load "
								+ serial
								+ " as certificate.");
					} else {
						System.out.println(
							"INF: " + serial + " is a certificate.");
						b = in.getSerialNumber();
					}
				}
				if (b == null) {
					System.out.println("ERR: Abort.");
					System.exit(1);
				} else {
					if (cert == null)
						cert = new PKICert(settings);
					if (data == null)
						data = new PKIDatabase(settings);
					if (cert.start()) {
						data.setCertifier(cert);
						if (data.start()) {
							cert.setDatabase(data);
							X509Certificate out = data.getCertificate(b);
							if (out != null) {
								System.out.println("Found: " + out);
							} else {
								System.out.println("No Person found.");
							}
							// Do the Query
						} else {
							System.out.println(
								"ERR: Database could not be started.");
						}
					} else {
						System.out.println(
							"ERR: Certifier could not be started.");
					}
				}
			} catch (Exception e) {
				System.out.println("ERR: Query failed: " + e);
				e.printStackTrace();
			}
	}

}
