/**
 * @version working
 * @author Thomas Kriegelstein
 * $Id: Sign.java,v 1.5 2003/12/16 08:36:58 tk4 Exp $
 */
package pki;

import java.io.File;
import java.io.FileInputStream;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.Properties;

/**
 * The sign class does command line person pseudonym certification.
 *
 * @author Thomas Kriegelstein
 * @version $Id: Sign.java,v 1.5 2003/12/16 08:36:58 tk4 Exp $
 */
public class Sign {
	/** settings for configuration of the PKI */
	private static Properties settings = new Properties();
	public static final String keyStoreFileName = "keyStoreFileName";
	public static final String secretKeyPassword = "secretKeyPassword";
	public static final String keyStorePassword = "keyStorePassword";
	public static final String dbType = "dbType";
	public static final String fileDBFileName = "fileDBFileName";
	public static final String mysqlHost = "mysqlHost";
	public static final String mysqlPort = "mysqlPort";
	public static final String mysqlDBName = "mysqlDBName";
	public static final String mysqlUsername = "mysqlUsername";
	public static final String mysqlPassword = "mysqlPassword";
	private static String settingsFileName = "data/pkiSettings.properties";

	static PKIDatabase data = null;
	static PKICert cert = null;
	static PKIServer serv = null;

	/**
	 * Prints a usage message to standard out
	 */
	private static void usage() {
		System.out.println("Usage: pki.Sign [options] command");
		String[] opts =
			{
				"Options are:",
				"  -keypass      %s  use secret for server operation",
				"  -storephrase  %s  use secret to validate the integrity of the keystore",
				"  -mysql            use mysql database instead of file database",
				"Commands are:",
				"  -sign         %s  replace certificate in file with a trusted one",
				"  -help             print this help message and exit" };
		for (int i = 0; i < opts.length; i++)
			System.out.println(opts[i]);
	}

	/**
	 * Reads settings from file
	 */
	private static void loadSettings() {
		try {
			File sf = new File(settingsFileName);
			settings.load(new FileInputStream(sf));
		} catch (Exception e) {
			System.out.println(
				"Unable to read Settings from file: " + settingsFileName);
		}
	}

	/**
	 * Sets default settings for the PKI Server
	 */
	private static void setDefaultSettings() {
		settings = new Properties();
		settings.setProperty(keyStoreFileName, "data/pkikeys");
		settings.setProperty(secretKeyPassword, "secret");
		settings.setProperty(keyStorePassword, "secret");
		settings.setProperty(dbType, "mysql");
		settings.setProperty(fileDBFileName, "data/pkidata");
		settings.setProperty(mysqlHost, "localhost");
		settings.setProperty(mysqlPort, "3306");
		settings.setProperty(mysqlDBName, "pki");
		settings.setProperty(mysqlUsername, "pkiadmin");
		settings.setProperty(mysqlPassword, "pkiadmin");
	}

	/**
	 * Main method. Handles commandline options.
	 */
	public static void main(String[] args) throws Exception {
		//System.out.println("Sign.main called.");
		setDefaultSettings();
		// load settings from file
		loadSettings();
		String filename = null;
		// get command line options
		for (int i = 0; i < args.length; i++) {
			if (args[i].equals("?") || args[i].equals("-help")) {
				usage();
				System.exit(0);
				break;
			} else if (args[i].equals("-storephrase")) {
				i += 1;
				if (i < args.length) {
					settings.setProperty(keyStorePassword, args[i]);
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else if (args[i].equals("-keypass")) {
				i += 1;
				if (i < args.length) {
					settings.setProperty(secretKeyPassword, args[i]);
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else if (args[i].equals("-mysql")) {
				settings.setProperty(dbType, "mysql");
			} else if (args[i].equals("-sign")) {
				i += 1;
				if (i < args.length) {
					filename = args[i];
				} else {
					usage();
					System.exit(-1);
					break;
				}
			} else {
				usage();
				System.exit(0);
				break;
			}
		}
		if (filename == null) {
			System.out.println("No command given.");
			usage();
			System.exit(-1);
		} else
			try {
				File f = new File(filename);
				if (!f.exists()) {
					System.out.println(
						"ERR: File: " + filename + " not found.");
				} else {
					X509Certificate in = PKIUtils.getFileCertificate(filename);
					if (in == null) {
						System.out.println("ERR: Could not load certificate");
					} else {
						if (cert == null)
							cert = new PKICert(settings);
						if (data == null)
							data = new PKIDatabase(settings);
						if (startCert()) {
							if (startDatabase()) {
								System.out.println("Signing: " + in);
								System.out.println(
									"If this is incorrect press Ctrl+C now, press Return otherwise.");
								System.in.read();
								Properties info = new Properties();
								info.putAll(PKIUtils.decrypt(in));
								PublicKey pk = in.getPublicKey();
								String pw =
									settings.getProperty(secretKeyPassword);
								X509Certificate out =
									cert.createPersonCertificate(
										pk,
										info,
										pw.toCharArray());
								System.out.println("Generated: " + out);
								PKIUtils.saveCertificate(filename, out);
							} else {
								System.out.println(
									"ERR: Database could not be started.");
							}
						} else {
							System.out.println(
								"ERR: Certifier could not be started.");
						}
					}
				}
			} catch (Exception e) {
				System.out.println("ERR: Sign failed: " + e);
				e.printStackTrace();
			}
	}

	/**
	 * Starts the certification object.
	 */
	public static boolean startCert() {
		System.out.println("Main.startCert called.");
		// trys to start the certifier
		if (!cert.start()) {
			return false;
		}
		data.setCertifier(cert);
		return true;
	}

	/**
	 * Starts the database object.
	 */
	public static boolean startDatabase() {
		System.out.println("Main.startDatabase called.");
		// Close old connection if neccessary
		if (data.isConnected()) {
			data.stop();
		}
		// trys to start the database connection
		if (!data.start()) {
			// notifies the gui about the result of starting
			return false;
		}
		cert.setDatabase(data);
		return true;
	}

}
